# ========================================================================
# Filter to implement an automatic or semi_automatic adjustment of:
# Luminosity, Contrast, Saturation
# ______________________________________________________________________
# initial version
 
# note: it includes a service function provided by Saul Goode (thanks!)
#       to detect the saturation level
 
# Flow implemented to get the final result:
# 
# 0.  Analyse action requested:
 
# A.  If action = Info only:
 
# A1. get from the histogram: the Mean Value, the Std Deviation, the Saturation level
# A2. compare them with the targets and calculate the deltas
# A3. according to these deltas classify:
#       the current Luminosity: Low, Average, High
#       the current Contrast: Low, Average, High
#       the current Saturation: Low, Average, High
# note: Average means around the target +/_ 32
# A4. display to the user the calculated classes
 
# B.  action = Enhance:
 
# B1. analyse the level of strength requested for the adjustments
# B2. if not "auto", all factors are equally set to the level requested
# B3. if "auto" each single factor is set according to the current class:
# B4.  weak for a low class
#      normal for an average class
#      weak for a high class
# B5. calculate the current Mean Value from the histogram
# B6. calculate the adjusting Set Value
# B7. apply it, using the gimp_levels native filter
# B8. calculate the current Contrast from the histogram (std deviation)
# B9. calculate the adjusting Set Value
# B10.apply it, using the gimp_brightess_contrast native filter
# B11.calculate the current Saturation from Saul Goode's service routine
# B12.calculate the adjusting Set Value
# B13.apply it, using the gimp_hue_saturation native filter 
# ________________________________________________________________
# ________________________________________________________________
# 
# This program is free software# you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation# either version 3 of the License, or
# (at your option) any later version.  
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY# without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program# if not, you can view the GNU General Public
# License version 3 at the web site http://www.gnu.org/licenses/gpl_3.0.html
# Alternatively you can write to the Free Software Foundation, Inc., 675 Mass
# Ave, Cambridge, MA 02139, USA.
 
#####################################################################

#!/usr/bin/env python
from gimpfu import * 
 
#####################################################################
   
def get_saturation (layer):

    image = pdb.gimp_item_get_image (layer)
    sat_image = pdb.plug_in_decompose_registered (image, layer, "Saturation", False )
    #decompose returns an array, in this case of 1 element
    sat_layer = pdb.gimp_image_get_active_layer (sat_image[0]) #element 0 is the only one returned
    saturation = pdb.gimp_histogram (sat_layer, HISTOGRAM_VALUE, 0, 255 )
    pdb.gimp_image_delete (sat_image[0])
    return(saturation[0])
  
# ===================================================================================================================
#  
def mod_LCS (inImage, inLayer, action, strength, target_L, target_C, target_S):

# Initialize an undo, so the process can be undone with a single undo
    pdb.gimp_image_undo_group_start (inImage)
    
# set-up constant values    
    LowL = 0
    AvgL = 1 
    HighL = 2
    LowC = 0
    AvgC = 1
    HighC = 2
    LowS = 0
    AvgS = 1
    HighS = 2
    LCSclass = ("Low", "Average", "High")
    RfactorL = 9
    RfactorC = 9
    RfactorS  = 9
    Luminosity = 9
    Contrast = 9
    Saturation = 9
  
    (meanValue, stdDev, median, pixels, count, percent) = pdb.gimp_histogram (inLayer, HISTOGRAM_VALUE, 0 ,255)
    
    meanSat = get_saturation (inLayer)  # mean of values
    
    if meanValue < (target_L - 32):
        Luminosity = LowL
    elif meanValue > (target_L + 32):
        Luminosity = HighL
    else:    
        Luminosity = AvgL 
         
    if stdDev < (target_C - 32):
        Contrast = LowC 
    elif stdDev > (target_C + 32):
        Contrast = HighC
    else:    
        Contrast = AvgC 
          
    if meanSat < (target_S - 32):
        Saturation = LowS 
    elif meanSat > (target_S + 32):
        Saturation = HighS
    else:    
        Saturation = AvgS
        
 # perform requested action
  
    if action == 1:    # logic for info only
        pdb.gimp_message ("mean value: "+str(meanValue))
        pdb.gimp_message ("std deviation: "+str(stdDev))
        pdb.gimp_message ("mean saturation: "+str(meanSat))
        pdb.gimp_message ("Luminosity class: "+LCSclass[Luminosity])
        pdb.gimp_message ("Contrast class: "+LCSclass[Contrast])
        pdb.gimp_message ("Saturation class: "+LCSclass[Saturation])

 # end if action = info only
 # =========================
 
 # logic for auto enhance start here
 # =================================
 
    if action == 0:    # logic for enhance
    
 # apply choosen strength

        if strength > 0:              # means: not auto
            if strength == 3:         # request for weak adjustment
                RfactorL = 3
            elif strength == 1:       # request for strong adjustmnet
                RfactorL = 1
            else:                     # request for normal adjustment
                RfactorL = 2      
      
            RfactorC = RfactorL       # all 3 equal
            RfactorS = RfactorL       # all 3 equal
     
        else:                         # auto adjustment <================================================
  # auto is based on current status
  
            if Luminosity == HighL:
                RfactorL = 3          # weak against high      
            elif Luminosity == LowL:
                RfactorL = 3          # weak against low
            else:    
                RfactorL = 2          # normal against average
  
            if Contrast == HighC:
                RfactorC = 3          # weak against high      
            elif Contrast == LowC:
                RfactorC = 3          # weak against low
            else:    
                RfactorC = 2          # normal against average
  
            if Saturation == HighS:
                RfactorS = 3          # weak against high      
            elif Saturation == LowS:
                RfactorS = 3          # weak against low
            else:    
                RfactorS = 2          # normal against average
         
        #pdb.gimp_message("RfL="+str(RfactorL)+", RfC="+str(RfactorC)+", RfS="+str(RfactorS))  

# mean value
     
        # avoid division by zero
        
        if meanValue < 0.1:            # check zero 
            meanValue = 0.1 
        
        setMeanValue = target_L / meanValue                           # the ratio desired/actual
        setMeanValue = (setMeanValue + RfactorL) / (RfactorL + 1)     # reduce the ratio
        setMeanValue = setMeanValue * setMeanValue                    # power of the ratio
        if setMeanValue < 0.1:                                        # check range: must be between 0.1 and 10
            setMeanValue = 0.1 
        elif setMeanValue > 10:
            setMeanValue = 10 
      
        #pdb.gimp_message("set midpoint: "+str (setMeanValue))
        
        pdb.gimp_levels (inLayer, HISTOGRAM_VALUE, 0, 255, setMeanValue, 0, 255)
        (newmeanValue, newstdDev, median, pixels, count, percent) = pdb.gimp_histogram (inLayer, HISTOGRAM_VALUE, 0 ,255)
        #pdb.gimp_message ("new mean value: "+str(newmeanValue))
    
# standard deviation
        setContrast = target_C - stdDev                                # the difference
        setContrast = setContrast / RfactorC                           # reduce the difference
        if setContrast < -127:                                         # check range: must be between -127 and 127 
            setContrast = -127 
        elif setContrast > 127:
            setContrast = 127 
          
        #pdb.gimp_message ("delta contrast: "+str(setContrast))
        
        pdb.gimp_brightness_contrast (inLayer, 0, setContrast)
        (newmeanValue, newstdDev, median, pixels, count, percent) = pdb.gimp_histogram (inLayer, HISTOGRAM_VALUE, 0 ,255)
        #pdb.gimp_message ("new mean value: "+str(newmeanValue))
        #pdb.gimp_message ("new std deviation: "+str(newstdDev))
     
# mean saturation    
        newmeanSat = get_saturation (inLayer)                           # actual sat (new, because may have been changed) 
        setSaturation = target_S - newmeanSat                           # the difference (using new)
        setSaturation = setSaturation / RfactorS                        # reduce the difference
        setSaturation = setSaturation * 100 / 128                       # set as percentage
        if setSaturation < -100:                                        # check range: must be between -100 and 100
            setSaturation = -100 
        elif setSaturation > 100:
            setSaturation = 100
         
        #pdb.gimp_message ("delta saturation: "+str(setSaturation))
        
        pdb.gimp_hue_saturation (inLayer, 0, 0, 0, setSaturation)
        newmeanSat = get_saturation (inLayer)                          
        #pdb.gimp_message ("new mean saturation: "+str(meanSat))
 

    pdb.gimp_displays_flush

    pdb.gimp_image_undo_group_end (inImage)
    
    return 
    
# This is the plugin registration function
register(
    "mod_LCS",    
    "to modify L, C, S values",   
    "This script modifies Luminosity, Contrast and Saturation according to target values",
    "Diego", 
    "Diego Nassetti ", 
    "December 2015",
    "mod_LCS", 
    "", 
    [
      (PF_IMAGE, "image", "Input Image", None),
      (PF_DRAWABLE, "drawable", "Input drawable", None),
      (PF_OPTION, "action", "Action requested", 0, ("enhance","info only")),
      (PF_OPTION, "strenght", "Strength of adjustment", 0, ("auto", "strong", "normal", "weak")),
      (PF_ADJUSTMENT, "target_L", "Target luminosity", 120, (32, 224, 8)),
      (PF_ADJUSTMENT, "target_C", "Target contrast", 80, (32, 224, 8)),
      (PF_ADJUSTMENT, "target_S", "Target saturation" , 120, (32, 224, 8)),
    ], 
    [],
    mod_LCS,
    menu="<Image>/Diego/Colours",
    )

main()
    